/// \file ijkgenscalar.cxx
/// generate a scalar field
/// Version v0.1.0

/*
  IJK: Isosurface Jeneration Code
  Copyright (C) 2011-2013 Rephael Wenger

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  (LGPL) as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <cmath>
#include <iomanip>
#include <iostream>
#include <vector>

#include "ijkNrrd.h"
#include "ijkgenscalar.h"
#include "ijkgenscalarIO.h"

#include "ijkgrid_nrrd.txx"
#include "ijkgradientfield.txx"
#include "ijkscalarfield.txx"
#include "ijkstring.txx"

// Types
typedef IJK::NRRD_DATA<int, int> NRRD_HEADER;

using namespace IJKSCALARFIELD;
using namespace IJKGENSCALAR;
using namespace std;

// global constants
const int DIM3(3);

// global variables
char * ofilename = NULL;
SET_VALUE<int> grid_axis_size;
vector<FIELD_INFO> field_info;
FIELD_PARAM field_param;
bool output_gradients = false;
bool flag_silent = false;

// functions to read field parameters
void init_field_info(vector<FIELD_INFO> & field_info);
void read_dimension();
void read_num_vert_per_axis();
void read_field_name(const vector<FIELD_INFO> & field_info);
void prompt_num_objects(const int field_index);

// misc functions
void construct_gradient_filename
(const char * scalar_filename, string & gradient_filename);
void parse_command_line(const int argc, char **argv);
void usage_error(), help_msg();
void check_input_param(const int dimension);
void check_field_options();

int main(int argc, char **argv)
{
  IJK::PROCEDURE_ERROR error("ijkgenscalar");

  try {

    init_field_info(field_info);

    parse_command_line(argc, argv);

    if (ofilename == NULL) {
      error.AddMessage("Missing output file name");
      throw error;
    }

    if (!field_param.field_index.IsSet()) 
      { read_field_name(field_info); }
    check_field_options();

    if (!field_param.dimension.IsSet()) 
      { read_dimension(); }

    check_input_param(field_param.Dimension());

    if (!grid_axis_size.IsSet())
      { read_num_vert_per_axis(); }
    field_param.SetAxisSize(grid_axis_size.Value());

    SCALAR_GRID scalar_grid
      (field_param.Dimension(), &(field_param.axis_size[0]));

    int ifield = field_param.FieldIndex();

    if (field_info[ifield].name == "dist2planes")
      { field_param.flag_multi = true; }

    if (!field_param.num_objects.IsSet()) {
      if (!field_param.flag_multi) 
        { field_param.num_objects.Set(1); }
      else {
        prompt_num_objects(field_param.FieldIndex());
      }
    }

    int dimension = field_param.Dimension();
    if (output_gradients) {
      GRADIENT_GRID gradient_grid;

      gradient_grid.SetSize
        (dimension, scalar_grid.AxisSize(), dimension);

      if (!field_info[ifield].IsGradientImplemented()) {
        error.AddMessage
          ("Programming error. Gradient not implemented for function ",
           field_info[ifield].name, ".");
        throw error;
      }

      field_info[ifield].gradient_function_ptr(scalar_grid, gradient_grid);

      if (!flag_silent) {
        cout << "Writing scalar field to " << ofilename << endl;
      }
      write_scalar_grid(ofilename, scalar_grid, field_param, field_info);

      string gradient_filename;
      construct_gradient_filename(ofilename, gradient_filename);
      if (!flag_silent) {
        cout << "Writing gradient field to " << gradient_filename << endl;
      }
      write_gradient_grid
        (gradient_filename, gradient_grid, field_param, field_info);
    }
    else {

      field_info[ifield].function_ptr(scalar_grid);
      if (field_param.flag_set_boundary2zero) {
        set_boundary_values(0, scalar_grid);
      }

      if (!flag_silent) {
        cout << "Writing scalar field to " << ofilename << endl;
      }
      write_scalar_grid(ofilename, scalar_grid, field_param, field_info);
    }

  } 
  catch (IJK::ERROR error) {
    if (error.NumMessages() == 0) {
      cerr << "Unknown error." << endl;
    }
    else { error.Print(cerr); }
    cerr << "Exiting." << endl;
    exit(20);
  }
  catch (...) {
    cerr << "Unknown error." << endl;
    exit(50);
  };

}

// **************************************************
// Prompt for scalar field parameters
// **************************************************

template <typename CTYPE>
void prompt_coordinates
(const char * prompt, const int num_coord,
 std::vector<CTYPE> & coord)
{
  cout << prompt << " (" << num_coord << " values) : ";
  for (int ic = 0; ic < num_coord; ic++) {
    CTYPE c;
    cin >> c;
    coord.push_back(c);
  }
}

template <typename CTYPE>
void prompt_coordinates
(const std::ostringstream & prompt, const int num_coord,
 std::vector<CTYPE> & coord)
{
  prompt_coordinates(prompt.str().c_str(), num_coord, coord);
}

/// Prompt for a single scalar value and add to list
template <typename STYPE>
void prompt_scalar
(const char * prompt, std::vector<STYPE> & list)
{
  cout << prompt << " : ";
  STYPE s;
  cin >> s;
  list.push_back(s);
}

void prompt_num_objects(const int field_index)
{
  string field_name = field_info[field_index].name;

  int n(0);
  while (true) {

    if (field_name == "cube") {
      cout << "Enter number of cubes: ";
    }
    else if (field_name == "cylinder") {
      cout << "Enter number of cylinders: ";
    }
    else if (field_name == "annulus") {
      cout << "Enter number of annuli: ";
    }
    else if (field_name == "flange") {
      cout << "Enter number of flanges: ";
    }
    else if (field_name == "dist2planes") {
      cout << "Enter number of planes: ";
    }
    else {
      cout << "Enter number of objects: ";
    }

    cin >> n;
    if (n > 0) { break; };

    cout << "Illegal number.  Enter a positive integer." << endl;
  }

  field_param.num_objects.Set(n);
}

void prompt_center
(const int dimension, const int num_obj, const char * obj_name)
{
  int k = field_param.center.size()/dimension;

  // Clear any stray coordinates
  if (field_param.center.size()%dimension != 0) 
    { field_param.center.resize(k*dimension); }

  if (k >= num_obj) { return; };

  if (num_obj == 1) {
    ostringstream prompt;
    prompt << "Enter coordinates of " << obj_name << " center";
    prompt_coordinates(prompt, dimension, field_param.center);
  }
  else {

    for (int i = k; i < num_obj; i++) {
      if (field_param.center.size() <= i*dimension) {
        ostringstream prompt;
        prompt << "Enter coordinates of " << obj_name << " " 
               << i << " center";
        prompt_coordinates(prompt, dimension, field_param.center);
      }
    }
  }
}

void prompt_two_axis_directions(const int dimension)
{
  if (field_param.direction.size() == 0) {
    prompt_coordinates("Enter first axis direction",
                       dimension, field_param.direction);
  }

  if (field_param.side_direction.size() == 0) {
    prompt_coordinates("Enter second axis direction",
                       dimension, field_param.side_direction);
  }
}

void prompt_translate(const int dimension)
{
  if (field_param.translate.size() == 0) {
    prompt_coordinates("Enter translation vector",
                       dimension, field_param.translate);
  }
}

void prompt_cylinder
(const int dimension, const int num_cylinders, const bool flag_closed)
{
  prompt_center(dimension, num_cylinders, "cylinder");

  if (field_param.flag_tilt) {
    if (field_param.direction.size() == 0) {
      prompt_coordinates("Enter cylinder axis direction",
                         dimension, field_param.direction);
    }
  }
  else {
    // Set axis direction to (1,0,0,...)
    field_param.direction.clear();
    field_param.direction.resize(dimension, 0);
    if (dimension > 0)
      { field_param.direction[0] = 1; }
  }

  if (flag_closed) {
    if (field_param.length_difference.size() == 0) {
      prompt_scalar("Enter difference between cylinder length and diameter", 
                    field_param.length_difference);
    }
  }

}

void prompt_cylinder_wedge(const int dimension)
{
  const int num_wedges = 1;

  prompt_cylinder(dimension, num_wedges, true);

  if (field_param.angle.size() == 0) {
    prompt_scalar("Enter wedge angle", field_param.angle);
  }

  if (field_param.side_direction.size() == 0) {
    prompt_coordinates
      ("Enter coordinates of side direction",
       dimension, field_param.side_direction);
  }

  if (field_param.center_scalar.size() == 0) {
    prompt_scalar("Enter scalar value at cylinder center", 
                  field_param.center_scalar);
  }

}

void prompt_dist2planes(const int dimension, const int num_planes)
{
  if (field_param.center.size() == 0) {
    prompt_coordinates("Enter coordinates of point on all planes",
                       dimension, field_param.center);
  }

  for (int i = 0; i < num_planes; i++) {
    ostringstream prompt;
    prompt << "Enter coordinates of normal to plane " << i;

    prompt_coordinates(prompt, dimension, field_param.normal);
  }
}

void prompt_edge(const int dimension)
{
  if (field_param.center.size() == 0) {
    prompt_coordinates("Enter coordinates of point on edge",
                       dimension, field_param.center);
  }

  if (field_param.angle.size() == 0) {
    prompt_scalar
      ("Enter dihedral angle", field_param.angle);
  }

  if (field_param.normal.size() == 0) {
    prompt_coordinates
      ("Enter coordinates of normal to first plane", 
       dimension, field_param.normal);
  }

  if (field_param.rotation_direction.size() == 0) {
    prompt_coordinates
      ("Enter rotation direction", dimension, field_param.rotation_direction);
  }
}

void prompt_square_cylinder(const int dimension)
{
  if (field_param.center.size() == 0) {
    prompt_coordinates("Enter coordinates of cylinder center",
                       dimension, field_param.center);
  }

  if (field_param.direction.size() == 0) {
    prompt_coordinates("Enter cylinder axis direction",
                       dimension, field_param.direction);
  }

  if (field_param.side_direction.size() == 0) {
    prompt_coordinates("Enter direction orthogonal to square side",
                       dimension, field_param.side_direction);
  }

  if (field_param.length_difference.size() == 0) {
    prompt_scalar("Enter difference between cylinder height and width", 
                  field_param.length_difference);
  }
}

void prompt_square_cylinder_rot45(const int dimension)
{
  if (field_param.center.size() == 0) {
    prompt_coordinates("Enter coordinates of cylinder center",
                       dimension, field_param.center);
  }

  if (field_param.direction.size() == 0) {
    prompt_coordinates("Enter cylinder axis direction",
                       dimension, field_param.direction);
  }

  if (field_param.side_direction.size() == 0) {
    prompt_coordinates("Enter square corner direction",
                       dimension, field_param.side_direction);
  }

  if (field_param.length_difference.size() == 0) {
    prompt_scalar("Enter difference between cylinder height and width", 
                  field_param.length_difference);
  }
}

void prompt_crossing_square_cylinders(const int dimension)
{
  if (field_param.center.size() == 0) {
    prompt_coordinates("Enter coordinates of crossing center",
                       dimension, field_param.center);
  }

  if (field_param.direction.size() == 0) {
    prompt_coordinates("Enter direction of first cylinder",
                       dimension, field_param.direction);
  }

  if (field_param.side_direction.size() == 0) {
    prompt_coordinates("Enter direction of second cylinder",
                       dimension, field_param.side_direction);
  }

  if (field_param.length_difference.size() == 0) {
    prompt_scalar("Enter difference between cylinder height and width", 
                  field_param.length_difference);
  }
}

void prompt_annulus(const int dimension)
{
  if (field_param.center.size() == 0) {
    prompt_coordinates("Enter coordinates of annulus center",
                       dimension, field_param.center);
  }

  if (field_param.direction.size() == 0) {
    prompt_coordinates("Enter annulus axis direction",
                       dimension, field_param.direction);
  }

  if (field_param.radius.size() == 0) {
    prompt_scalar
      ("Enter annulus radius", field_param.radius);
  }

  if (field_param.length_difference.size() == 0) {
    prompt_scalar("Enter difference between annulus height and width", 
                  field_param.length_difference);
  }

}

void prompt_flange(const int dimension)
{
  if (field_param.center.size() == 0) {
    prompt_coordinates("Enter coordinates of flange center",
                       dimension, field_param.center);
  }

  if (field_param.direction.size() == 0) {
    prompt_coordinates("Enter flange axis direction",
                       dimension, field_param.direction);
  }

  if (field_param.radius.size() == 0) {
    prompt_scalar
      ("Enter flange radius", field_param.radius);
  }

  if (field_param.length_difference.size() == 0) {
    prompt_scalar("Enter difference between flange height and width", 
                  field_param.length_difference);
  }

}

template <typename T>
void prompt_maxval(IJKGENSCALAR::SET_VALUE<T> & maxval)
{
  T x;

  cout << "Enter maximum scalar value: ";
  cin >> x;
  maxval.Set(x);
}

template <typename T>
void prompt_random_seed(IJKGENSCALAR::SET_VALUE<T> & random_seed)
{
  T x;

  cout << "Enter random seed: ";
  cin >> x;
  random_seed.Set(x);
}

// **************************************************
// Generate scalar fields
// **************************************************

void gen_cube(SCALAR_GRID & grid)
{
  const int dimension = grid.Dimension();
  const int num_cubes = field_param.NumObjects();
  IJK::PROCEDURE_ERROR error("gen_cube");

  if (field_param.flag_tilt && dimension != 3) {
    error.AddMessage("Programming error.  Illegal dimension ", dimension,
                     " used with tilt.");
    error.AddMessage("  Tilt can only be used with dimension 3.");
    throw error;
  }

  SCALAR_GRID gridB;
  gridB.SetSize(grid);

  if (field_param.flag_stack) {
    prompt_center(dimension, 1, "first cube");
    prompt_translate(dimension);
  }
  else {
    prompt_center(dimension, num_cubes, "cube");
  }

  if (field_param.flag_tilt) {
    IJK::ARRAY<COORD_TYPE> xdir(dimension), ydir(dimension), zdir(dimension);

    prompt_two_axis_directions(dimension);

    compute_orthogonal_basis3D
      (field_param.DirectionPtrConst(0), field_param.SideDirectionPtrConst(0),
       zdir.Ptr(), xdir.Ptr(), ydir.Ptr());

    if (field_param.spacing.size() >= dimension) {
      gen_cube
        (field_param.CenterPtrConst(0), 
         xdir.PtrConst(), ydir.PtrConst(), zdir.PtrConst(),
         field_param.SpacingPtrConst(), grid);
    }
    else {
      gen_cube
        (field_param.CenterPtrConst(0), 
         xdir.PtrConst(), ydir.PtrConst(), zdir.PtrConst(), grid);
    }

    if (field_param.flag_stack) {
      IJK::ARRAY<COORD_TYPE> translate(dimension);
      IJK::ARRAY<COORD_TYPE> cube_center(dimension);

      transform_basis3D
        (field_param.TranslatePtrConst(0), 
         xdir.PtrConst(), ydir.PtrConst(), zdir.PtrConst(), translate.Ptr());

      for (int i = 1; i < num_cubes; i++) {
        IJK::multiply_coord
          (dimension, i, translate.PtrConst(), cube_center.Ptr());
        IJK::add_coord(dimension, field_param.CenterPtrConst(0), 
                       cube_center.Ptr(), cube_center.Ptr());

        if (field_param.spacing.size() >= dimension) {
          gen_cube
            (cube_center.PtrConst(),
             xdir.PtrConst(), ydir.PtrConst(), zdir.PtrConst(),
             field_param.SpacingPtrConst(), gridB);
        }
        else {
          gen_cube
            (cube_center.PtrConst(),
             xdir.PtrConst(), ydir.PtrConst(), zdir.PtrConst(), gridB);
        }
        min_scalar(grid, gridB, grid);
      }

    }
    else {
      for (int i = 1; i < num_cubes; i++) {

        if (field_param.spacing.size() >= dimension) {
          gen_cube
            (field_param.CenterPtrConst(i), 
             xdir.PtrConst(), ydir.PtrConst(), zdir.PtrConst(),
             field_param.SpacingPtrConst(), gridB);
        }
        else {
          gen_cube
            (field_param.CenterPtrConst(i), 
             xdir.PtrConst(), ydir.PtrConst(), zdir.PtrConst(), gridB);
        }
        min_scalar(grid, gridB, grid);
      }
    }

    if (field_param.spacing.size() >= dimension) 
      { grid.SetSpacing(field_param.SpacingPtrConst()); }

  }
  else {

    gen_dist2point_Linf(field_param.CenterPtrConst(0), grid);

    if (field_param.flag_stack) {
      const GRADIENT_COORD_TYPE * translate = field_param.TranslatePtrConst(0);
      IJK::ARRAY<COORD_TYPE> cube_center(dimension);

      for (int i = 1; i < num_cubes; i++) {
        IJK::multiply_coord(dimension, i, translate, cube_center.Ptr());
        IJK::add_coord(dimension, field_param.CenterPtrConst(0), 
                       cube_center.Ptr(), cube_center.Ptr());
                     
        gen_dist2point_Linf(cube_center.Ptr(), gridB); 
        min_scalar(grid, gridB, grid);
      }
    }
    else {

      for (int i = 1; i < num_cubes; i++) {
        gen_dist2point_Linf(field_param.CenterPtrConst(i), gridB); 
        min_scalar(grid, gridB, grid);
      }
    }

  }
}

void gen_gradient_cube(SCALAR_GRID & scalar_grid, GRADIENT_GRID & gradient_grid)
{
  const int dimension = scalar_grid.Dimension();
  const int num_cubes = field_param.NumObjects();
  const bool gradient_discontinuity_zero =
    field_param.gradient_discontinuity_zero;
  SCALAR_GRID gridB;
  GRADIENT_GRID gradientB;
  IJK::PROCEDURE_ERROR error("gen_gradient_cube");

  if (field_param.flag_tilt && dimension != 3) {
    error.AddMessage("Programming error.  Illegal dimension ", dimension,
                     " used with tilt.");
    error.AddMessage("  Tilt can only be used with dimension 3.");
    throw error;
  }

  gridB.SetSize(scalar_grid);
  gradientB.SetSize(gradient_grid);

  if (field_param.flag_stack) {
    prompt_center(dimension, 1, "first cube");
    prompt_translate(dimension);
  }
  else {
    prompt_center(dimension, num_cubes, "cube");
  }

  if (field_param.flag_tilt) {
    IJK::ARRAY<COORD_TYPE> xdir(dimension), ydir(dimension), zdir(dimension);

    prompt_two_axis_directions(dimension);

    compute_orthogonal_basis3D
      (field_param.DirectionPtrConst(0), field_param.SideDirectionPtrConst(0),
       zdir.Ptr(), xdir.Ptr(), ydir.Ptr());

    if (field_param.spacing.size() >= dimension) {
      gen_gradient_cube
        (field_param.CenterPtrConst(0), 
         xdir.PtrConst(), ydir.PtrConst(), zdir.PtrConst(),
         field_param.SpacingPtrConst(), scalar_grid, gradient_grid);
    }
    else {
      gen_gradient_cube
        (field_param.CenterPtrConst(0), 
         xdir.PtrConst(), ydir.PtrConst(), zdir.PtrConst(),
         scalar_grid, gradient_grid);
    }

    if (field_param.flag_stack) {
      IJK::ARRAY<COORD_TYPE> translate(dimension);
      IJK::ARRAY<COORD_TYPE> cube_center(dimension);

      transform_basis3D
        (field_param.TranslatePtrConst(0), 
         xdir.PtrConst(), ydir.PtrConst(), zdir.PtrConst(), translate.Ptr());

      for (int i = 1; i < num_cubes; i++) {
        IJK::multiply_coord
          (dimension, i, translate.PtrConst(), cube_center.Ptr());
        IJK::add_coord(dimension, field_param.CenterPtrConst(0), 
                       cube_center.Ptr(), cube_center.Ptr());

        if (field_param.spacing.size() >= dimension) {
          gen_gradient_cube
            (cube_center.Ptr(),
             xdir.PtrConst(), ydir.PtrConst(), zdir.PtrConst(),
             field_param.SpacingPtrConst(), gridB, gradientB);
        }
        else {
          gen_gradient_cube
            (cube_center.Ptr(),
             xdir.PtrConst(), ydir.PtrConst(), zdir.PtrConst(),
             gridB, gradientB);
        }

        min_scalar_select_gradient
          (scalar_grid, gradient_grid, gridB, gradientB, 
           gradient_discontinuity_zero, scalar_grid, gradient_grid);
      }
    }
    else {

      for (int i = 1; i < num_cubes; i++) {

        if (field_param.spacing.size() >= dimension) {
          gen_gradient_cube
            (field_param.CenterPtrConst(i),
             xdir.PtrConst(), ydir.PtrConst(), zdir.PtrConst(),
             field_param.SpacingPtrConst(), gridB, gradientB);
        }
        else {
          gen_gradient_cube
            (field_param.CenterPtrConst(i),
             xdir.PtrConst(), ydir.PtrConst(), zdir.PtrConst(),
             gridB, gradientB);
        }

        min_scalar_select_gradient
          (scalar_grid, gradient_grid, gridB, gradientB, 
           gradient_discontinuity_zero, scalar_grid, gradient_grid);
      }
    }

    if (field_param.spacing.size() >= dimension) {
      scalar_grid.SetSpacing(field_param.SpacingPtrConst()); 
      gradient_grid.SetSpacing(field_param.SpacingPtrConst()); 
    }
  }
  else {

    gen_gradient_dist2point_Linf
      (field_param.CenterPtrConst(0), gradient_discontinuity_zero,
       scalar_grid, gradient_grid);

    if (field_param.flag_stack) {
      const GRADIENT_COORD_TYPE * translate = field_param.TranslatePtrConst(0);
      IJK::ARRAY<COORD_TYPE> cube_center(dimension);

      for (int i = 1; i < num_cubes; i++) {
        IJK::multiply_coord(dimension, i, translate, cube_center.Ptr());
        IJK::add_coord(dimension, field_param.CenterPtrConst(0), 
                       cube_center.Ptr(), cube_center.Ptr());

        gen_gradient_dist2point_Linf
          (cube_center.Ptr(), gradient_discontinuity_zero,
           gridB, gradientB); 

        min_scalar_select_gradient
          (scalar_grid, gradient_grid, gridB, gradientB, 
           gradient_discontinuity_zero, scalar_grid, gradient_grid);
      }
    }
    else {

      for (int i = 1; i < num_cubes; i++) {
        gen_gradient_dist2point_Linf
          (field_param.CenterPtrConst(i), gradient_discontinuity_zero,
           gridB, gradientB); 

        min_scalar_select_gradient
          (scalar_grid, gradient_grid, gridB, gradientB, 
           gradient_discontinuity_zero, scalar_grid, gradient_grid);
      }
    }

  }
}

void gen_closed_cylinder(SCALAR_GRID & grid)
{
  const int dimension = grid.Dimension();
  const int num_cylinders = field_param.NumObjects();
  SCALAR_GRID gridB;

  if (dimension < 1) { return; }

  gridB.SetSize(grid);

  prompt_cylinder(dimension, num_cylinders, true);

  gen_closed_cylinder
    (field_param.CenterPtrConst(0), field_param.DirectionPtrConst(0),
     field_param.length_difference[0], grid);

  for (int i = 1; i < num_cylinders; i++) {
    gen_closed_cylinder
      (field_param.CenterPtrConst(i), field_param.DirectionPtrConst(0),
       field_param.length_difference[0], gridB);

    min_scalar(grid, gridB, grid);
  }
}

void gen_gradient_closed_cylinder
(SCALAR_GRID & scalar_grid, GRADIENT_GRID & gradient_grid)
{
  const int dimension = scalar_grid.Dimension();
  const int num_cylinders = field_param.NumObjects();
  SCALAR_GRID gridB;
  GRADIENT_GRID gradientB;

  gridB.SetSize(scalar_grid);
  gradientB.SetSize(gradient_grid);

  prompt_cylinder(dimension, num_cylinders, true);

  gen_gradient_closed_cylinder
    (field_param.CenterPtrConst(0), field_param.DirectionPtrConst(0),
     field_param.length_difference[0], scalar_grid, gradient_grid);

  for (int i = 1; i < num_cylinders; i++) {
    gen_gradient_closed_cylinder
      (field_param.CenterPtrConst(i), field_param.DirectionPtrConst(0),
       field_param.length_difference[0], gridB, gradientB);

    min_scalar_select_gradient
      (scalar_grid, gradient_grid, gridB, gradientB, 
       field_param.gradient_discontinuity_zero, scalar_grid, gradient_grid);
  }
}


/// Compute a direction at a given angle from \a dir1.
/// @param dir1[] Direction.
/// @param dir2[] Second direction.
/// @param angle Angle in degrees.
/// @param[out] new_dir[] Direction at given with \a dir1 in plane
///                  determined by \a dir1 and \a dir2.
/// @pre dir1[] is a unit vector.
/// @pre dir2[] is a unit vector.
template <typename COORD1_TYPE, typename COORD2_TYPE, typename COORD3_TYPE>
void compute_dir_at_angle
(const int dimension, const COORD1_TYPE dir1[], 
 const COORD2_TYPE dir2[], const ANGLE_TYPE angle,
 COORD3_TYPE new_dir[])
{
  const double radians = M_PI*angle/180.0;

  IJK::multiply_coord(dimension, cos(radians), dir1, new_dir);
  IJK::add_scaled_coord
    (dimension, sin(radians), dir2, new_dir, new_dir);
}

template <typename T>
T ijk_mod(const T x, const int y)
{
  if (x >= 0) {
    return(remainder(x, y));
  }
  else {
    return(y-remainder(-x, y));
  }
}


/// Compute two normal vectors to wedge planes.
/// @param[out] normal[] Normals to wedge planes.
template <typename CTYPE>
void compute_wedge_plane_normals(CTYPE normal[2*DIM3])
{
  IJK::ARRAY<COORD_TYPE> xdir(DIM3), ydir(DIM3), zdir(DIM3);

  compute_orthogonal_basis3D
    (field_param.DirectionPtrConst(0), field_param.SideDirectionPtrConst(0), 
     zdir.Ptr(), xdir.Ptr(), ydir.Ptr());

  IJK::copy_coord(DIM3, xdir.PtrConst(), normal);

  if (ijk_mod(field_param.angle[0], 360) == 180) {
    IJK::copy_coord(DIM3, xdir.PtrConst(), normal+DIM3);
  }
  else {
    compute_dir_at_angle
      (DIM3, xdir.PtrConst(), ydir.PtrConst(),
       180-field_param.angle[0], normal+DIM3);
  }
}

void gen_cylinder_wedge(SCALAR_GRID & grid)
{
  const int dimension = grid.Dimension();
  double normal[2*DIM3];
  SCALAR_GRID gridB;
  const int num_planes = 2;
  IJK::PROCEDURE_ERROR error("gen_cylinder_wedge");

  if (dimension < 1) { return; }

  if (dimension != 3) {
    error.AddMessage("Programming error.  Illegal dimension ", dimension,
                     ".");
    error.AddMessage("  Cylinder wedge can only be used with dimension 3.");
    throw error;
  }

  prompt_cylinder_wedge(dimension);

  compute_wedge_plane_normals(normal);

  gridB.SetSize(grid);

  gen_closed_cylinder
    (field_param.CenterPtrConst(0), field_param.DirectionPtrConst(0),
     field_param.length_difference[0], grid);


  if (ijk_mod(field_param.angle[0], 360) <= 180) {
    gen_max_dist2planes
      (field_param.CenterPtrConst(0), normal, num_planes, gridB);
  }
  else {
    gen_min_dist2planes
      (field_param.CenterPtrConst(0), normal, num_planes, gridB);
  }
  gridB.Add(field_param.center_scalar[0]);

  max_scalar(grid, gridB, grid);
}

void gen_gradient_cylinder_wedge
(SCALAR_GRID & grid, GRADIENT_GRID & gradient)
{
  const int dimension = grid.Dimension();
  double normal[2*DIM3];
  SCALAR_GRID gridB;
  GRADIENT_GRID gradientB;
  const int num_planes = 2;
  IJK::PROCEDURE_ERROR error("gen_cylinder_wedge");

  if (dimension < 1) { return; }

  if (dimension != 3) {
    error.AddMessage("Programming error.  Illegal dimension ", dimension,
                     ".");
    error.AddMessage("  Cylinder wedge can only be used with dimension 3.");
    throw error;
  }

  prompt_cylinder_wedge(dimension);

  compute_wedge_plane_normals(normal);

  gridB.SetSize(grid);
  gradientB.SetSize(gradient);

  gen_gradient_closed_cylinder
    (field_param.CenterPtrConst(0), field_param.DirectionPtrConst(0),
     field_param.length_difference[0], grid, gradient);


  if (ijk_mod(field_param.angle[0], 360) <= 180) {
    gen_gradient_max_dist2planes
      (field_param.CenterPtrConst(0), normal, num_planes, gridB, gradientB);
  }
  else {
    gen_gradient_min_dist2planes
      (field_param.CenterPtrConst(0), normal, num_planes, gridB, gradientB);
  }
  gridB.Add(field_param.center_scalar[0]);

  max_scalar_select_gradient
    (grid, gradient, gridB, gradientB, 
     field_param.gradient_discontinuity_zero, grid, gradient);
}

void gen_octahedron(SCALAR_GRID & grid)
{
  const int dimension = grid.Dimension();

  if (field_param.center.size() == 0) {
    prompt_coordinates("Enter coordinates of octahedron center",
                       dimension, field_param.center);
  }

  gen_dist2point_L1(field_param.CenterPtrConst(0), grid);
}

void gen_gradient_octahedron
(SCALAR_GRID & scalar_grid, GRADIENT_GRID & gradient_grid)
{
  const int dimension = scalar_grid.Dimension();

  if (field_param.center.size() == 0) {
    prompt_coordinates("Enter coordinates of octahedron center",
                       dimension, field_param.center);
  }

  if (field_param.gradient_discontinuity_zero) {
    gen_gradient_dist2point_L1_zero
      (field_param.CenterPtrConst(0), scalar_grid, gradient_grid);
  }
  else {
    gen_gradient_dist2point_L1_select
      (field_param.CenterPtrConst(0), scalar_grid, gradient_grid);
  }
}

void gen_sfield_with_constant_unit_gradient(SCALAR_GRID & grid)
{
  const int dimension = grid.Dimension();

  if (field_param.center.size() == 0) {
    prompt_coordinates("Enter coordinates of point with scalar value zero",
                       dimension, field_param.center);
  }

  if (field_param.direction.size() == 0) {
    prompt_coordinates("Enter gradient direction",
                       dimension, field_param.direction);
  }

  gen_constant_unit_gradient
    (field_param.CenterPtrConst(0), field_param.DirectionPtrConst(0), grid);
}

void gen_gfield_with_constant_unit_gradient
(SCALAR_GRID & scalar_grid, GRADIENT_GRID & gradient_grid)
{
  const int dimension = scalar_grid.Dimension();

  if (field_param.center.size() == 0) {
    prompt_coordinates("Enter coordinates of point with scalar value zero",
                       dimension, field_param.center);
  }

  if (field_param.direction.size() == 0) {
    prompt_coordinates("Enter gradient direction",
                       dimension, field_param.direction);
  }

  gen_constant_unit_gradient
    (field_param.CenterPtrConst(0), field_param.DirectionPtrConst(0), 
     scalar_grid, gradient_grid);
}

void gen_open_cylinder(SCALAR_GRID & grid)
{
  const int dimension = grid.Dimension();
  const int num_cylinders = field_param.NumObjects();

  prompt_cylinder(dimension, num_cylinders, false);

  gen_dist2line_L2
    (field_param.CenterPtrConst(0), field_param.DirectionPtrConst(0), 
     grid);

}

void gen_gradient_open_cylinder
(SCALAR_GRID & scalar_grid, GRADIENT_GRID & gradient_grid)
{
  const int dimension = scalar_grid.Dimension();
  const int num_cylinders = field_param.NumObjects();

  prompt_cylinder(dimension, num_cylinders, false);

  gen_gradient_dist2line_L2
    (field_param.CenterPtrConst(0), field_param.DirectionPtrConst(0), 
     scalar_grid, gradient_grid);
}

void gen_square_cylinder(SCALAR_GRID & grid)
{
  const int dimension = grid.Dimension();
  IJK::ARRAY<COORD_TYPE> xdir(dimension), ydir(dimension), zdir(dimension);

  prompt_square_cylinder(dimension);

  compute_orthogonal_basis3D
    (field_param.DirectionPtrConst(0), field_param.SideDirectionPtrConst(0),
     zdir.Ptr(), xdir.Ptr(), ydir.Ptr());

  gen_closed_square_cylinder
    (field_param.CenterPtrConst(0), 
     xdir.PtrConst(), ydir.PtrConst(), zdir.PtrConst(), 
     field_param.length_difference[0], grid);
}

void gen_gradient_square_cylinder
(SCALAR_GRID & scalar_grid, GRADIENT_GRID & gradient_grid)
{
  const int dimension = scalar_grid.Dimension();
  IJK::ARRAY<COORD_TYPE> xdir(dimension), ydir(dimension), zdir(dimension);

  prompt_square_cylinder(dimension);

  compute_orthogonal_basis3D
    (field_param.DirectionPtrConst(0), field_param.SideDirectionPtrConst(0),
     zdir.Ptr(), xdir.Ptr(), ydir.Ptr());

  gen_gradient_closed_square_cylinder
    (field_param.CenterPtrConst(0), 
     xdir.PtrConst(), ydir.PtrConst(), zdir.PtrConst(), 
     field_param.length_difference[0], scalar_grid, gradient_grid);
}

void gen_square_cylinder_rot45(SCALAR_GRID & grid)
{
  const int dimension = grid.Dimension();
  IJK::ARRAY<COORD_TYPE> xdir(dimension), ydir(dimension), zdir(dimension);

  prompt_square_cylinder_rot45(dimension);

  compute_orthogonal_basis3D
    (field_param.DirectionPtrConst(0), field_param.SideDirectionPtrConst(0),
     zdir.Ptr(), xdir.Ptr(), ydir.Ptr());

  gen_closed_square_cylinder_rot45
    (field_param.CenterPtrConst(0), 
     xdir.PtrConst(), ydir.PtrConst(), zdir.PtrConst(), 
     field_param.length_difference[0], grid);
}

void gen_gradient_square_cylinder_rot45
(SCALAR_GRID & scalar_grid, GRADIENT_GRID & gradient_grid)
{
  const int dimension = scalar_grid.Dimension();
  IJK::ARRAY<COORD_TYPE> xdir(dimension), ydir(dimension), zdir(dimension);

  prompt_square_cylinder_rot45(dimension);

  compute_orthogonal_basis3D
    (field_param.DirectionPtrConst(0), field_param.SideDirectionPtrConst(0),
     zdir.Ptr(), xdir.Ptr(), ydir.Ptr());

  gen_gradient_closed_square_cylinder_rot45
    (field_param.CenterPtrConst(0), 
     xdir.PtrConst(), ydir.PtrConst(), zdir.PtrConst(), 
     field_param.length_difference[0], scalar_grid, gradient_grid);
}

void gen_crossing_square_cylinders(SCALAR_GRID & grid)
{
  const int dimension = grid.Dimension();
  IJK::ARRAY<COORD_TYPE> dir0(dimension);
  IJK::ARRAY<COORD_TYPE> dir1(dimension);
  IJK::ARRAY<COORD_TYPE> dir2(dimension);

  prompt_crossing_square_cylinders(dimension);

  compute_orthogonal_basis3D
    (field_param.DirectionPtrConst(0), field_param.SideDirectionPtrConst(0),
     dir0.Ptr(), dir1.Ptr(), dir2.Ptr());
	
  SCALAR_GRID gridA, gridB, gridC;
  gridA.SetSize(grid);
  gridB.SetSize(grid);
  gridC.SetSize(grid);

  gen_closed_square_cylinder
    (field_param.CenterPtrConst(0), 
     dir0.PtrConst(), dir1.PtrConst(), dir2.PtrConst(),
     field_param.length_difference[0], gridA);
  gen_closed_square_cylinder
    (field_param.CenterPtrConst(0), 
     dir1.PtrConst(), dir2.PtrConst(), dir0.PtrConst(),
     field_param.length_difference[0], gridB);
  gen_closed_square_cylinder
    (field_param.CenterPtrConst(0), 
     dir0.PtrConst(), dir2.PtrConst(), dir1.PtrConst(),
     field_param.length_difference[0], gridC);

  min_scalar(gridA, gridB, grid);
  min_scalar(gridC, grid, grid);
}

void gen_gradient_crossing_square_cylinders
(SCALAR_GRID & scalar_grid, GRADIENT_GRID & gradient_grid)
{
  const int dimension = scalar_grid.Dimension();
  IJK::ARRAY<COORD_TYPE> dir0(dimension);
  IJK::ARRAY<COORD_TYPE> dir1(dimension);
  IJK::ARRAY<COORD_TYPE> dir2(dimension);

  prompt_crossing_square_cylinders(dimension);

  compute_orthogonal_basis3D
    (field_param.DirectionPtrConst(0), field_param.SideDirectionPtrConst(0),
     dir0.Ptr(), dir1.Ptr(), dir2.Ptr());
	
  SCALAR_GRID gridA, gridB, gridC;
  GRADIENT_GRID gradientA, gradientB, gradientC;
  gridA.SetSize(scalar_grid);
  gridB.SetSize(scalar_grid);
  gridC.SetSize(scalar_grid);
  gradientA.SetSize(gradient_grid);
  gradientB.SetSize(gradient_grid);
  gradientC.SetSize(gradient_grid);

  gen_gradient_closed_square_cylinder
    (field_param.CenterPtrConst(0), 
     dir0.PtrConst(), dir1.PtrConst(), dir2.PtrConst(),
     field_param.length_difference[0], gridA, gradientA);
  gen_gradient_closed_square_cylinder
    (field_param.CenterPtrConst(0), 
     dir1.PtrConst(), dir2.PtrConst(), dir0.PtrConst(),
     field_param.length_difference[0], gridB, gradientB);
  gen_gradient_closed_square_cylinder
    (field_param.CenterPtrConst(0), 
     dir0.PtrConst(), dir2.PtrConst(), dir1.PtrConst(),
     field_param.length_difference[0], gridC, gradientC);

  min_scalar_select_gradient
    (gridA, gradientA, gridB, gradientB, 
     field_param.gradient_discontinuity_zero, scalar_grid, gradient_grid);
  min_scalar_select_gradient
    (scalar_grid, gradient_grid, gridC, gradientC,
     field_param.gradient_discontinuity_zero, scalar_grid, gradient_grid);
}

void gen_crossing_square45_cylinders(SCALAR_GRID & grid)
{
  const int dimension = grid.Dimension();
  IJK::ARRAY<COORD_TYPE> origin(dimension);
  IJK::ARRAY<COORD_TYPE> dir0(dimension);
  IJK::ARRAY<COORD_TYPE> dir1(dimension);
  IJK::ARRAY<COORD_TYPE> dir2(dimension);

  prompt_crossing_square_cylinders(dimension);

  compute_orthogonal_basis3D
    (field_param.DirectionPtrConst(0), field_param.SideDirectionPtrConst(0),
     dir0.Ptr(), dir1.Ptr(), dir2.Ptr());
	
  SCALAR_GRID gridA, gridB, gridC;
  gridA.SetSize(grid);
  gridB.SetSize(grid);
  gridC.SetSize(grid);

  gen_closed_square_cylinder_rot45
    (field_param.CenterPtrConst(0), 
     dir0.PtrConst(), dir1.PtrConst(), dir2.PtrConst(),
     field_param.length_difference[0], gridA);
  gen_closed_square_cylinder_rot45
    (field_param.CenterPtrConst(0), 
     dir1.PtrConst(), dir2.PtrConst(), dir0.PtrConst(),
     field_param.length_difference[0], gridB);
  gen_closed_square_cylinder_rot45
    (field_param.CenterPtrConst(0), 
     dir0.PtrConst(), dir2.PtrConst(), dir1.PtrConst(),
     field_param.length_difference[0], gridC);

  min_scalar(gridA, gridB, grid);
  min_scalar(gridC, grid, grid);
}

void gen_gradient_crossing_square45_cylinders
(SCALAR_GRID & scalar_grid, GRADIENT_GRID & gradient_grid)
{
  const int dimension = scalar_grid.Dimension();
  IJK::ARRAY<COORD_TYPE> dir0(dimension);
  IJK::ARRAY<COORD_TYPE> dir1(dimension);
  IJK::ARRAY<COORD_TYPE> dir2(dimension);

  prompt_crossing_square_cylinders(dimension);

  compute_orthogonal_basis3D
    (field_param.DirectionPtrConst(0), field_param.SideDirectionPtrConst(0),
     dir0.Ptr(), dir1.Ptr(), dir2.Ptr());
	
  SCALAR_GRID gridA, gridB, gridC;
  GRADIENT_GRID gradientA, gradientB, gradientC;
  gridA.SetSize(scalar_grid);
  gridB.SetSize(scalar_grid);
  gridC.SetSize(scalar_grid);
  gradientA.SetSize(gradient_grid);
  gradientB.SetSize(gradient_grid);
  gradientC.SetSize(gradient_grid);

  gen_gradient_closed_square_cylinder_rot45
    (field_param.CenterPtrConst(0), 
     dir0.PtrConst(), dir1.PtrConst(), dir2.PtrConst(),
     field_param.length_difference[0], gridA, gradientA);
  gen_gradient_closed_square_cylinder_rot45
    (field_param.CenterPtrConst(0), 
     dir1.PtrConst(), dir2.PtrConst(), dir0.PtrConst(),
     field_param.length_difference[0], gridB, gradientB);
  gen_gradient_closed_square_cylinder_rot45
    (field_param.CenterPtrConst(0), 
     dir0.PtrConst(), dir2.PtrConst(), dir1.PtrConst(),
     field_param.length_difference[0], gridC, gradientC);

  min_scalar_select_gradient
    (gridA, gradientA, gridB, gradientB, 
     field_param.gradient_discontinuity_zero, scalar_grid, gradient_grid);
  min_scalar_select_gradient
    (scalar_grid, gradient_grid, gridC, gradientC,
     field_param.gradient_discontinuity_zero, scalar_grid, gradient_grid);
}

void gen_annulus(SCALAR_GRID & grid)
{
  const int dimension = grid.Dimension();

  prompt_annulus(dimension);

  gen_annulus
    (field_param.CenterPtrConst(0), field_param.DirectionPtrConst(0),
     field_param.radius[0], field_param.length_difference[0], grid);
}

void gen_gradient_annulus(SCALAR_GRID & grid, GRADIENT_GRID & gradient)
{
  const int dimension = grid.Dimension();

  prompt_annulus(dimension);

  gen_gradient_annulus
    (field_param.CenterPtrConst(0), field_param.DirectionPtrConst(0),
     field_param.radius[0], field_param.length_difference[0], 
     grid, gradient);
}

void gen_flange(SCALAR_GRID & grid)
{
  const int dimension = grid.Dimension();

  prompt_flange(dimension);

  gen_flange(field_param.CenterPtrConst(0), field_param.DirectionPtrConst(0),
             field_param.radius[0], field_param.length_difference[0], grid);
}

void gen_gradient_flange(SCALAR_GRID & grid, GRADIENT_GRID & gradient)
{
  const int dimension = grid.Dimension();

  prompt_flange(dimension);

  gen_gradient_flange
    (field_param.CenterPtrConst(0), field_param.DirectionPtrConst(0),
     field_param.radius[0], field_param.length_difference[0], grid, gradient);
}

void gen_torus(SCALAR_GRID & grid)
{
  const int dimension = grid.Dimension();
  const int num_tori = field_param.NumObjects();

  SCALAR_GRID gridB;
  gridB.SetSize(grid);

  prompt_center(dimension, num_tori, "torus");

  if (field_param.direction.size() == 0) {
    prompt_coordinates("Enter torus axis direction",
                       dimension, field_param.direction);
  }

  if (field_param.radius.size() == 0) {
    prompt_scalar
      ("Enter torus radius", field_param.radius);
  }

  gen_torus
    (field_param.CenterPtrConst(0), field_param.DirectionPtrConst(0),
     field_param.radius[0], grid);

  for (int i = 1; i < num_tori; i++) {
    gen_torus
      (field_param.CenterPtrConst(i), field_param.DirectionPtrConst(0),
       field_param.radius[0], gridB);
    min_scalar(grid, gridB, grid);
  }

}

void gen_gradient_torus(SCALAR_GRID & grid, GRADIENT_GRID & gradient)
{
  const int dimension = grid.Dimension();
  const int num_tori = field_param.NumObjects();

  SCALAR_GRID gridB;
  GRADIENT_GRID gradientB;
  gridB.SetSize(grid);
  gradientB.SetSize(gradient);

  prompt_center(dimension, num_tori, "torus");

  if (field_param.direction.size() == 0) {
    prompt_coordinates("Enter torus axis direction",
                       dimension, field_param.direction);
  }

  if (field_param.radius.size() == 0) {
    prompt_scalar
      ("Enter torus radius", field_param.radius);
  }

  gen_gradient_torus
    (field_param.CenterPtrConst(0), field_param.DirectionPtrConst(0),
     field_param.radius[0], grid, gradient);

  for (int i = 1; i < num_tori; i++) {
    gen_gradient_torus
      (field_param.CenterPtrConst(i), field_param.DirectionPtrConst(0),
       field_param.radius[0], gridB, gradientB);

    min_scalar_select_gradient
      (grid, gradient, gridB, gradientB, 
       field_param.gradient_discontinuity_zero, grid, gradient);
  }

}

void gen_dist2planes(SCALAR_GRID & scalar_grid)
{
  const int dimension = scalar_grid.Dimension();
  const int num_planes = field_param.NumObjects();

  prompt_dist2planes(dimension, num_planes);

  gen_max_dist2planes(field_param.CenterPtrConst(0),
                      field_param.NormalPtrConst(), num_planes,
                      scalar_grid);
}

void gen_gradient_dist2planes
(SCALAR_GRID & scalar_grid, GRADIENT_GRID & gradient)
{
  const int dimension = scalar_grid.Dimension();
  const int num_planes = field_param.NumObjects();

  prompt_dist2planes(dimension, num_planes);

  gen_gradient_max_dist2planes
    (field_param.CenterPtrConst(0), field_param.NormalPtrConst(), num_planes,
     scalar_grid, gradient);
}

void gen_edge(SCALAR_GRID & scalar_grid)
{
  const int dimension = scalar_grid.Dimension();
  const int num_planes = 2;
  double normal[2*dimension];
  double v2[dimension];

  prompt_edge(dimension);

  IJK::copy_coord(dimension, field_param.NormalPtrConst(), normal);
  normalize_vector(dimension, normal);

  compute_normalized_orthogonal_vector
    (dimension, field_param.RotationDirectionPtrConst(), normal, v2);

  compute_dir_at_angle
    (dimension, normal, v2, field_param.angle[0], normal+dimension);

  gen_max_dist2planes
    (field_param.CenterPtrConst(0), normal, num_planes, scalar_grid);
}

void gen_gradient_edge(SCALAR_GRID & scalar_grid, GRADIENT_GRID & gradient)
{
  const int dimension = scalar_grid.Dimension();
  const int num_planes = 2;
  double normal[2*dimension];
  double v2[dimension];

  prompt_edge(dimension);

  IJK::copy_coord(dimension, field_param.NormalPtrConst(), normal);
  normalize_vector(dimension, normal);

  compute_normalized_orthogonal_vector
    (dimension, field_param.RotationDirectionPtrConst(), normal, v2);

  compute_dir_at_angle
    (dimension, normal, v2, field_param.angle[0], normal+dimension);

  gen_gradient_max_dist2planes
    (field_param.CenterPtrConst(0), normal, num_planes, scalar_grid, gradient);
}

void gen_random_int(SCALAR_GRID & grid)
{
  if (!field_param.maxval.IsSet()) 
    { prompt_maxval(field_param.maxval); }

  if (!field_param.random_seed.IsSet())
    { prompt_random_seed(field_param.random_seed); }

  gen_random_int(field_param.MaxVal(), field_param.RandomSeed(), grid);
}


// **************************************************
// Read functions
// **************************************************

void read_dimension()
{
  int dimension;

  cout << "Enter domain dimension: ";
  cin >> dimension;

  field_param.SetDimension(dimension);

  if (dimension < 1) {
    cerr << "Illegal dimension.  Dimension must be at least 1." << endl;
    exit(20);
  };
}

void read_num_vert_per_axis()
{
  int size;

  cout << "Enter # grid vertices per axis: ";
  cin >> size;

  grid_axis_size.Set(size);

  if (size < 2) {
    cerr << "Illegal grid axis size.  Must be at least 2." << endl;
    exit(30);
  };

}

// **************************************************
// FIELD_INFO functions
// **************************************************

void init_field_info(vector<FIELD_INFO> & field_info)
{
  FIELD_INFO info;

  info.Set("cube", gen_cube, gen_gradient_cube);
  info.SetFlagCenter(true);
  info.SetDescription
    ("L-infinity distance to a point. Isosurfaces are cubes.");
  field_info.push_back(info);

  info.Set("cylinder", gen_closed_cylinder, gen_gradient_closed_cylinder);
  info.SetDescription
    ("Minimum of L2 distance to a line and an orthogonal plane.");
  info.AddToDescription("Isosurfaces are closed cylinders.");
  field_info.push_back(info);

  info.Set("cylinder_wedge", gen_cylinder_wedge, gen_gradient_cylinder_wedge);
  info.SetDescription("Isosurfaces are cylinder wedges.");
  field_info.push_back(info);

  info.Set("annulus", gen_annulus, gen_gradient_annulus);
  info.SetAnnulusFlags();
  info.SetDescription
    ("Minimum of L2 distance to a cylinder and an orthogonal plane.");
  info.AddToDescription("Isosurfaces are thickened annuli.");
  field_info.push_back(info);

  info.SetCylinderFlags();
  info.Set("torus", gen_torus, gen_gradient_torus);
  info.SetDescription("L2 distance to a circle.  Isosurfaces are torii.");
  field_info.push_back(info);

  info.Set("flange", gen_flange, gen_gradient_flange);
  info.SetAnnulusFlags();
  info.SetDescription("Isosurfaces are flanges.");
  field_info.push_back(info);

  info.Set("edge", gen_edge, gen_gradient_edge);
  info.SetFlagCenter(true);
  info.SetDescription
    ("Scalar field is the maximum signed distance to two planes.");
  info.AddToDescription
    ("Isosurfaces have a single edge with the given dihedral angle.");
  field_info.push_back(info);

  info.Set("dist2planes", gen_dist2planes, gen_gradient_dist2planes);
  info.SetFlagCenter(true);
  info.SetDescription
    ("Scalar field is the maximum signed distance to a set of planes.");
  field_info.push_back(info);

  info.Set("octahedron", gen_octahedron, gen_gradient_octahedron);
  info.SetFlagCenter(true);
  info.SetDescription("Isosurfaces are octahedra.");
  field_info.push_back(info);

  info.Set("square", gen_square_cylinder, gen_gradient_square_cylinder);
  info.SetCylinderFlags();
  info.SetDescription("Isosurfaces are closed square cylinders.");
  field_info.push_back(info);

  info.Set("square45", gen_square_cylinder_rot45,
           gen_gradient_square_cylinder_rot45);
  info.SetDescription("Isosurfaces are closed square cylinders");
  info.AddToDescription("  rotated 45 degrees around their central axes.");
  field_info.push_back(info);

  info.Set("squareX", gen_crossing_square_cylinders, 
           gen_gradient_crossing_square_cylinders);
  info.SetDescription
    ("Isosurfaces are 3 intersecting closed square cylinders.");
  field_info.push_back(info);

  info.Set("square45X", gen_crossing_square45_cylinders,
           gen_gradient_crossing_square45_cylinders);
  info.SetDescription
    ("Isosurfaces are 3 intersecting closed square cylinders.");
  info.AddToDescription("Square cylinders are rotated 45 degrees"); 
  info.AddToDescription("  around their central axes.");
  field_info.push_back(info);

  info.Set("constant_unit_gradient", 
           gen_sfield_with_constant_unit_gradient,
           gen_gfield_with_constant_unit_gradient);
  info.SetFlagCenter(true);
  info.SetDescription("Field with constant unit gradient.");
  info.AddToDescription("Isosurfaces are planes orthogonal to gradient.");
  field_info.push_back(info);

  info.Set("open_cylinder", gen_open_cylinder, gen_gradient_open_cylinder);
  info.SetCylinderFlags();
  info.SetDescription
    ("L2 distance to a line.  Isosurfaces are open cylinders.");
  field_info.push_back(info);

  info.Set("randomint", gen_random_int);
  info.SetRandomFlags();
  info.SetDescription
    ("Assign random integers as scalar values at grid vertices.");
  field_info.push_back(info);
}

bool find_field_name
(const vector<FIELD_INFO> & field_info, const string & s, int & ifield)
{
  ifield = 0;
  for (int i = 0; i < field_info.size(); i++) {
    if (field_info[i].name == s) {
      ifield = i;
      return(true);
    }
  }

  return(false);
}

/// Output field names.
/// @param with_gradients 
///           If true, output only fields with gradients implemented.
void out_field_names
(const vector<FIELD_INFO> & field_info, 
 const bool with_gradients)
{
  bool newline_flag(false);

  int iout = 0;
  for (int i = 0; i < field_info.size(); i++) {
    if (!with_gradients || 
        field_info[i].IsGradientImplemented()) {
      if (iout > 0) { 
        cout << ","; 
        newline_flag = false;
        if (iout%5 == 0) { 
          newline_flag = true;
          cout << endl; 
        }
      }
      cout << field_info[i].name;
      iout++;
    }
  }

  if (!newline_flag) { cout << endl; }
}

/// Output field descriptions.
void out_field_descriptions
(const vector<FIELD_INFO> & field_info)
{
  const int COL1_WIDTH = 15;

  for (int i = 0; i < field_info.size(); i++) {
    int num_description_lines = field_info[i].description.size();
    string s = field_info[i].name + ":";

    cout << left << setw(COL1_WIDTH) << s;

    int j = 0;
    if ((num_description_lines > 0) && (s.size()+1 <= COL1_WIDTH)) {
      cout << field_info[i].description[0];
      j++;
    }
    cout << endl;

    while (j < num_description_lines) {
      cout << setw(COL1_WIDTH) << " " 
           << field_info[i].description[j] << endl;
      j++;
    }
  }

}

void read_field_name(const vector<FIELD_INFO> & field_info)
{
  string s;
  int ifield;

  cout << "Enter scalar field name (h for help): ";
  cin >> s;

  while (true) {
    if (!find_field_name(field_info, s, ifield)) {
      if (s != "h" && s != "help") {
        cout << "Illegal field name: " << s << endl;
      }
    }
    else if (output_gradients &&
             !field_info[ifield].IsGradientImplemented()) {
      cout << "Computing gradients not implemented for function " 
             << field_info[ifield].name << "." << endl;
    }
    else {
      break;
    }

    cout << endl;
    cout << "Valid field names: ";
    out_field_names(field_info, output_gradients);
    cout << endl;
    cout << "Enter scalar field name (h for help): ";
    cin >> s;
  }

  field_param.SetFieldIndex(ifield);
}

// **************************************************
// STRING PROCESSING
// **************************************************

void split_string(const string & s, const char c,
                  string & prefix, string & suffix)
// split string at last occurrence of character c into prefix and suffix
{
  string::size_type i = s.rfind(c);
  if (i == string::npos) {
    prefix = s;
    suffix = "";
  }
  else {
    if (i > 0) { prefix = s.substr(0,i); }
    else { prefix = ""; };

    if (i+1 < s.length()) { suffix = s.substr(i+1, s.length()-i-1); }
    else { suffix = ""; };
  }
}

void construct_gradient_filename
(const char * scalar_filename, string & gradient_filename)
{
  string prefix, suffix;

  gradient_filename = scalar_filename;

  split_string(gradient_filename, '.', prefix, suffix);
  if (suffix == "nrrd" || suffix == "nhdr") { 
    gradient_filename = prefix; 
    gradient_filename = gradient_filename + ".grad." + suffix;
  }
  else {
    gradient_filename = gradient_filename + ".grad";
  }
}


// **************************************************
// COMMAND LINE PARSER
// **************************************************

void check_input_param(const int dimension)
{
  if (field_param.center.size()%dimension != 0) {
    cerr << "Error."
         << " Number of arguments in option -center does not match dimension "
         << dimension << ".";
    exit(50);
  }

  if (field_param.direction.size()%dimension != 0) {
    cerr << "Error."
         << " Number of arguments in option -dir does not match dimension "
         << dimension << ".";
    exit(50);
  }

  if (field_param.flag_tilt && dimension != 3) {
    cerr << "Error."
         << " Illegal dimension " << dimension << " used with tilt."
         << endl;
    cerr << "  Tilt can only be used with dimension 3." << endl;
    exit(50);
  }

}

void check_field_options()
{
  int ifield = field_param.FieldIndex();

  if (output_gradients) {

    if (!field_info[ifield].IsGradientImplemented()) {
      cerr << "Usage error.  Option -grad cannot be used with field "
           << field_info[ifield].name << "." << endl;
      usage_error();
    }
  }

}

int get_int(const int iarg, const int argc, char **argv)
{
  if (iarg+1 >= argc) { 
    cerr << "Usage error. Missing argument for option " 
         << argv[iarg] << " and missing file name." << endl;
    usage_error(); 
  }

  int x;
  if (!IJK::string2val(argv[iarg+1], x)) {
    cerr << "Error in argument for option: " << argv[iarg] << endl;
    cerr << "Non-integer character in string: " << argv[iarg+1] << endl;
    exit(50);
  }

  return(x);
}

float get_float(const int iarg, const int argc, char **argv)
{
  if (iarg+1 >= argc) { 
    cerr << "Usage error. Missing argument for option " 
         << argv[iarg] << " and missing file name." << endl;
    usage_error(); 
  }

  float x;
  if (!IJK::string2val(argv[iarg+1], x)) {
    cerr << "Error in argument for option: " << argv[iarg] << endl;
    cerr << "Non-numeric character in string: " << argv[iarg+1] << endl;
    exit(50);
  }

  return(x);
}

/// Get argument for option argv[iarg].
/// Does not modify iarg.
template <typename ETYPE>
void get_float
(const int iarg, const int argc, char **argv, vector<ETYPE> & v)
{
  if (iarg+1 >= argc) { 
    cerr << "Usage error. Missing argument for option " 
         << argv[iarg] << " and missing file name." << endl;
    usage_error();
  }

  if (!IJK::string2vector(argv[iarg+1], v)) {
    cerr << "Error in argument for option: " << argv[iarg] << endl;
    cerr << "Non-numeric character in string: " << argv[iarg+1] << endl;
    exit(50);
  }
}

int get_field_index
(const int iarg, const int argc, char **argv)
{
  int ifield(0);

  if (iarg+1 >= argc) { 
    cerr << "Usage error. Missing argument for option " 
         << argv[iarg] << " and missing file name." << endl;
    usage_error();
  }

  if (!find_field_name(field_info, argv[iarg+1], ifield)) {
    cerr << "Illegal field name: " << argv[iarg+1] << endl;
    exit(51);
  }

  return(ifield);
}

void parse_command_line(const int argc, char **argv)
{
  int iarg = 1;
  while (iarg < argc && argv[iarg][0] == '-') {

    string s = argv[iarg];

    if (s == "-dim") {
      int dimension = get_int(iarg, argc, argv);
      field_param.SetDimension(dimension);
      iarg++;
    }
    else if (s == "-asize") {
      int size = get_int(iarg, argc, argv);
      grid_axis_size.Set(size);
      iarg++;
    }
    else if (s == "-field") {
      int ifield = get_field_index(iarg, argc, argv);
      field_param.SetFieldIndex(ifield);
      iarg++;
    }
    else if (s == "-grad") {
      output_gradients = true;
    }
    else if (s == "-disc_zero") {
      field_param.gradient_discontinuity_zero = true;
    }
    else if (s == "-disc_select") {
      field_param.gradient_discontinuity_zero = false;
    }
    else if (s == "-tilt") {
      field_param.flag_tilt = true;
    }
    else if (s == "-no_tilt") {
      field_param.flag_tilt = false;
    }
    else if (s == "-multi") {
      field_param.flag_multi = true;
    }
    else if (s == "-single") {
      field_param.flag_multi = false;
    }
    else if (s == "-two") {
      field_param.flag_multi = true;
      field_param.num_objects.Set(2);
    }
    else if (s == "-n") {
      int num_obj = get_int(iarg, argc, argv);
      field_param.flag_multi = true;
      field_param.num_objects.Set(num_obj);
      iarg++;
    }
    else if (s == "-stack") {
      field_param.flag_stack = true;
    }
    else if (s == "-center") {
      get_float(iarg, argc, argv, field_param.center);
      iarg++;
    }
    else if (s == "-dir") {
      get_float(iarg, argc, argv, field_param.direction);
      iarg++;
    }
    else if (s == "-side_dir") {
      get_float(iarg, argc, argv, field_param.side_direction);
      iarg++;
    }
    else if (s == "-normal") {
      get_float(iarg, argc, argv, field_param.normal);
      iarg++;
    }
    else if (s == "-rot_dir") {
      get_float(iarg, argc, argv, field_param.rotation_direction);
      iarg++;
    }
    else if (s == "-spacing") {
      get_float(iarg, argc, argv, field_param.spacing);
      iarg++;
    }
    else if (s == "-translate") {
      get_float(iarg, argc, argv, field_param.translate);
      iarg++;
    }
    else if (s == "-radius") {
      float x = get_float(iarg, argc, argv);
      field_param.radius.push_back(x);
      iarg++;
    }
    else if (s == "-length_diff") {
      float x = get_float(iarg, argc, argv);
      field_param.length_difference.push_back(x);
      iarg++;
    }
    else if (s == "-angle") {
      ANGLE_TYPE angle = get_float(iarg, argc, argv);
      field_param.angle.push_back(angle);
      iarg++;
    }
    else if (s == "-center_scalar") {
      SCALAR_TYPE center_scalar = get_float(iarg, argc, argv);
      field_param.center_scalar.push_back(center_scalar);
      iarg++;
    }
    else if (s == "-seed") {
      int x = get_int(iarg, argc, argv);
      field_param.random_seed.Set(x);
      iarg++;
    }
    else if (s == "-maxval") {
      int x = get_int(iarg, argc, argv);
      field_param.maxval.Set(x);
      iarg++;
    }
    else if (s == "-bzero") {
      field_param.flag_set_boundary2zero = true;
    }
    else if (s == "-list") {
      cout << "Valid field names: " << endl;
      out_field_descriptions(field_info);
      exit(20);
    }
    else if (s == "-s") {
      flag_silent = true;
    }
    else if (s == "-help") {
      help_msg();
    }
    else {
      cerr << "Usage error.  Illegal parameter: " << s << endl;
      usage_error();
    };
    iarg++;
  };

  if (iarg+1 > argc) { 
    cerr << "Error.  Missing file name." << endl << endl;
    usage_error(); 
  }
  else if (iarg+1 < argc) 
    { usage_error(); }

  ofilename = argv[iarg];
}

void usage_msg()
{
  cerr << "Usage: ijkgenscalar [OPTIONS] <filename>" << endl;
  cerr << "OPTIONS:" << endl;
  cerr << "  [-field <field name>] [-grad]" << endl;
  cerr << "  [-dim <dimension>] [-asize <N>] [-tilt | -no_tilt]" << endl;
  cerr << "  [-single | -two | -multi | -n <num_obj>] [-stack]" << endl;
  cerr << "  [-center \"<coord>\"] [-dir \"<coord>\"]" 
       << endl;
  cerr << "  [-side_dir \"<coord>\"] [-rot_dir \"<coord>\"]" << endl;
  cerr << "  [-normal \"<coord>\"] [-translate \"<coord>\"]" << endl;
  cerr << "  [-radius <x>] [-length_diff <x>]" << endl;
  cerr << "  [-disc_zero] [-disc_select]" << endl;

  /* NOT FULLY IMPLEMENTED
  cerr << "  [-spacing \"<space_coord>\"]" << endl;
  */

  cerr << "  [-seed <S>] [-bzero]" << endl;
  cerr << "  [-s] [-list] [-help]" << endl;
}

void usage_error()
{
  usage_msg();
  exit(10);
}

void help_msg()
{
  cerr << "Usage: ijkgenscalar [OPTIONS] <filename>" << endl;
  cerr << "OPTIONS:" << endl;
  cerr << "  -field <field name>: Set field."
       << "  (Use -list to list fields.)" << endl;
  cerr << "  -grad: Write gradients to file." << endl;
  cerr << "  -dim <dimension>: Set dimension." << endl;
  cerr << "  -asize <N>: Set number of vertices along each grid axis to <N>."
       << endl;
  cerr << "  -tilt:    Use directions to tilt objects." << endl;
  cerr << "  -no_tilt: Objects are not tilted." << endl; 
  cerr << "  -single:  Single object (default)." << endl;
  cerr << "  -two:     Two objects." << endl;
  cerr << "  -multi:   Multiple objects." << endl;
  cerr << "  -n <num_obj>:   Number of objects is <num_obj>." << endl;
  cerr << "  -stack:   Stack of translated objects." << endl;
  cerr << "  -center \"<coord>\": Set center to given coordinates." << endl;
  cerr << "  -dir \"<coord>\": Set direction to given coordinates." << endl;
  cerr << "  -start_dir \"<coord>\": Set start direction to given coordinates." 
       << endl;
  cerr << "  -side_dir \"<coord>\": Set side direction to given coordinates." 
       << endl;
  cerr << "  -rot_dir \"<coord>\": Set rotation direction to given coordinates."
       << endl;
  cerr << "  -normal \"<coord>\": Set plane normals to given coordinates." 
       << endl;
  cerr << "  -translate \"<coord>\": Set stack translation vector to given coordinates." 
       << endl;
  cerr << "  -radius <x>: Set radius to <x>." << endl;
  cerr << "  -length_diff <x>:  Set (length-width) or (height-width) to <x>."
       << endl;
  cerr << "            Negative <x> sets width greater than length or height."
       << endl;
  cerr << "  -disc_zero: Set gradient to zero at gradient discontinuity." 
       << endl;
  cerr << "  -disc_select: Select gradient from one incident surface" << endl;
  cerr << "                at gradient discontinuity." << endl;

  /* NOT FULLY IMPLEMENTED
  cerr << "  -spacing \"<space_coord>\": Set spacing between grid vertices"
       << endl
       << "                along each axis." << endl;
  */

  cerr << "  -seed <S>: Set random seed to <S>." << endl;
  cerr << "  -bzero: Set values at all boundary vertices to zero." << endl;
  cerr << "  -list: List fields." << endl;
  cerr << "  -s: Silent mode." << endl;
  cerr << "  -help: Print this help message." << endl;

  exit(20);
}

