/// \file ijkgenscalar2.h
/// generate a scalar field

/*
  IJK: Isosurface Jeneration Kode
  Copyright (C) 2011-2013 Rephael Wenger

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  (LGPL) as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef _IJKGENSCALAR_TXX_
#define _IJKGENSCALAR_TXX_

#include <sstream>
#include <string>
#include <vector>

#include "ijk.txx"

namespace IJKGENSCALAR {

  // ********************************************************
  // TEMPLATE CLASS FIELD_INFO_T
  // ********************************************************

  /// Information about a field.
  /// Field name, scalar field generator, gradient field generator, etc.
  template <typename SCALAR_GRID_TYPE, typename GRADIENT_GRID_TYPE>
  class FIELD_INFO_T {

  protected:
    void Init();

  public:
    typedef void (*FUNCTION_PTR)(SCALAR_GRID_TYPE & grid);
    typedef void (*GRADIENT_FUNCTION_PTR)
      (SCALAR_GRID_TYPE & scalar_grid, GRADIENT_GRID_TYPE & gradient_grid);

  public:
    std::string name;
    std::vector<std::string> description;
    FUNCTION_PTR function_ptr;
    GRADIENT_FUNCTION_PTR gradient_function_ptr;
    bool flag_center;           ///< If true, field has a center.
    bool flag_direction;        ///< If true, field has a direction.
    bool flag_radius;           ///< If true, field has a radius.
    bool flag_length_difference;  ///< If true, field has length difference.
    bool flag_cylinder;         ///< If true, field has cylinder properties.
    bool flag_translate;        ///< If true, field has a translation vector.
    bool flag_random_seed;      ///< If true, field has a random seed.
    bool flag_maxval;           ///< If true, field has a maximum scalar value.

  public:
    FIELD_INFO_T() { Init(); };

    void Set(const std::string & field_name, FUNCTION_PTR f_ptr)
    {
      name = field_name;
      function_ptr = f_ptr;
      gradient_function_ptr = NULL;
      SetAllFlags(false);
    }

    void Set(const std::string & field_name, FUNCTION_PTR f_ptr,
             GRADIENT_FUNCTION_PTR grad_f_ptr)
    {
      name = field_name;
      function_ptr = f_ptr;
      gradient_function_ptr = grad_f_ptr;
      SetAllFlags(false);
    }

    void AddToDescription(const char * s)
    { description.push_back(s); }

    void SetDescription(const char * s)
    {
      description.clear();
      AddToDescription(s);
    }

    void SetFlagCenter(const bool flag_center);
    void SetCylinderFlags();
    void SetAnnulusFlags();
    void SetRandomFlags();
    void SetAllFlags(const bool flag);

    /// Return true if gradient_function_ptr is not NULL.
    bool IsGradientImplemented() const
    {
      if (gradient_function_ptr == NULL) 
        { return(false); }
      else
        { return(true); }
    }

  };

  // ********************************************************
  // TEMPLATE CLASS FIELD_PARAM_T
  // ********************************************************

  // forward declaration of template class SET_VALUE
  template <typename T> class SET_VALUE;

  /// Input parameters determining field properties.
    template <typename DIM_TYPE, typename AXIS_SIZE_TYPE,
              typename COORD_TYPE, typename DIR_TYPE,
              typename RADIUS_TYPE, typename DIFF_TYPE, typename ANGLE_TYPE,
              typename SCALAR_TYPE,
              typename SEED_TYPE,
              typename MIN_MAX_TYPE, typename NUM_TYPE>
  class FIELD_PARAM_T {

  protected:
    void Init();

  public:
    FIELD_PARAM_T(){ Init(); };

    SET_VALUE<DIM_TYPE> dimension;
    SET_VALUE<NUM_TYPE> field_index;     ///< Index into field info[].
    SET_VALUE<NUM_TYPE> num_objects;     ///< Number of cubes, tori, etc.
    SET_VALUE<SEED_TYPE> random_seed;    ///< Random seed.
    SET_VALUE<MIN_MAX_TYPE> maxval;      ///< Max scalar value.

    std::vector<AXIS_SIZE_TYPE> axis_size;
    std::vector<COORD_TYPE> center;
    std::vector<DIR_TYPE> direction;
    std::vector<DIR_TYPE> side_direction;
    std::vector<DIR_TYPE> rotation_direction;
    std::vector<DIR_TYPE> translate;
    std::vector<DIR_TYPE> normal;
    std::vector<RADIUS_TYPE> radius;
    std::vector<DIFF_TYPE> length_difference;
    std::vector<ANGLE_TYPE> angle;
    std::vector<SCALAR_TYPE> center_scalar;
    std::vector<COORD_TYPE> spacing;
    bool flag_tilt;
    bool flag_multi;
    bool flag_stack;
    bool gradient_discontinuity_zero;

    ///< If true, set all boundary vertices to zero.
    bool flag_set_boundary2zero;

    // Set functions
    void SetDimension(const DIM_TYPE dimension)
    { this->dimension.Set(dimension); }
    void SetFieldIndex(const NUM_TYPE field_index)
    { this->field_index.Set(field_index); }
    void SetAxisSize(const AXIS_SIZE_TYPE size);
    
    // get functions
    DIM_TYPE Dimension() const
    { return(dimension.Value()); }
    NUM_TYPE FieldIndex() const
    { return(field_index.Value()); }
    NUM_TYPE NumObjects() const
    { return(num_objects.Value()); }
    NUM_TYPE NumCenters() const
    { return(center.size()/Dimension()); }
    NUM_TYPE NumDirections() const
    { return(direction.size()/Dimension()); }
    NUM_TYPE NumRadii() const
    { return(radius.size()); }
    NUM_TYPE NumLengthDifferences() const
    { return(length_difference.size()); }
    NUM_TYPE NumSideDirections() const
    { return(side_direction.size()/Dimension()); }
    NUM_TYPE NumTranslate() const
    { return(translate.size()/Dimension()); }
    NUM_TYPE NumNormals() const
    { return(normal.size()/Dimension()); }
    NUM_TYPE NumRotationDirections() const
    { return(rotation_direction.size()/Dimension()); }
    NUM_TYPE NumAngles() const
    { return(angle.size()); }
    const SEED_TYPE RandomSeed() const
    { return(random_seed.Value()); }
    MIN_MAX_TYPE MaxVal() const
    { return(maxval.Value()); }
    const COORD_TYPE * CenterPtrConst(const NUM_TYPE i) const
      { return(&(center[0]) + i*Dimension()); }
    const COORD_TYPE * DirectionPtrConst(const NUM_TYPE i) const
      { return(&(direction[0]) + i*Dimension()); }
    const COORD_TYPE * SideDirectionPtrConst(const NUM_TYPE i) const
      { return(&(side_direction[0]) + i*Dimension()); }
    const COORD_TYPE * RotationDirectionPtrConst() const
      { return(&(rotation_direction[0])); }
    const COORD_TYPE * TranslatePtrConst(const NUM_TYPE i) const
      { return(&(translate[0]) + i*Dimension()); }
    const COORD_TYPE * SpacingPtrConst() const
      { return(&(spacing[0])); }
    const COORD_TYPE * NormalPtrConst() const
      { return(&(normal[0])); }

    void GetCenterString(const NUM_TYPE i, std::string & s) const;
    void GetDirectionString(const NUM_TYPE i, std::string & s) const;
    void GetSideDirectionString(const NUM_TYPE i, std::string & s) const;
    void GetRotationDirectionString(const NUM_TYPE i, std::string & s) const;
    void GetNormalString(const NUM_TYPE i, std::string & s) const;
    void GetTranslateString(const NUM_TYPE i, std::string & s) const;
    void GetRadiusString(const NUM_TYPE i, std::string & s) const;
    void GetAngleString(const NUM_TYPE i, std::string & s) const;
    void GetLengthDifferenceString(const NUM_TYPE i, std::string & s) const;
  };

  // ********************************************************
  // TEMPLATE CLASS SET_VALUE
  // ********************************************************

  template <typename T>
  class SET_VALUE {

  protected:
    T v;
    bool is_set;

    void Init() 
    { is_set = false; }

    void Init(const T value)
    {
      v = value;
      is_set = false;
    }

  public:
    SET_VALUE(){ Init(); };
    SET_VALUE(const T value) { Init(value); };

    void Set(const T value)
    {
      v = value;
      is_set = true;
    }

    bool IsSet() const { return(is_set); };
    T Value() const { return(v); };
  };

  // ********************************************************
  // TEMPLATE CLASS FIELD_INFO MEMBER FUNCTIONS
  // ********************************************************

  template <typename SCALAR_GRID_TYPE, typename GRADIENT_GRID_TYPE>
    void FIELD_INFO_T<SCALAR_GRID_TYPE,GRADIENT_GRID_TYPE>::
    Init()
    {
      SetAllFlags(false);
    }

  template <typename SCALAR_GRID_TYPE, typename GRADIENT_GRID_TYPE>
    void FIELD_INFO_T<SCALAR_GRID_TYPE,GRADIENT_GRID_TYPE>::
    SetFlagCenter(const bool flag_center)
    { 
      this->flag_center = flag_center;
    }

  template <typename SCALAR_GRID_TYPE, typename GRADIENT_GRID_TYPE>
    void FIELD_INFO_T<SCALAR_GRID_TYPE,GRADIENT_GRID_TYPE>::
    SetCylinderFlags()
    { 
      flag_center = true;
      flag_direction = true;
      flag_length_difference = true;
      flag_cylinder = true;
    }

  template <typename SCALAR_GRID_TYPE, typename GRADIENT_GRID_TYPE>
    void FIELD_INFO_T<SCALAR_GRID_TYPE,GRADIENT_GRID_TYPE>::
    SetAnnulusFlags()
    { 
      flag_center = true;
      flag_direction = true;
      flag_radius = true;
      flag_length_difference = true;
   }

  template <typename SCALAR_GRID_TYPE, typename GRADIENT_GRID_TYPE>
    void FIELD_INFO_T<SCALAR_GRID_TYPE,GRADIENT_GRID_TYPE>::
    SetRandomFlags()
    {
      flag_random_seed = true;
      flag_maxval = true;
    }

  template <typename SCALAR_GRID_TYPE, typename GRADIENT_GRID_TYPE>
    void FIELD_INFO_T<SCALAR_GRID_TYPE,GRADIENT_GRID_TYPE>::
    SetAllFlags(const bool flag)
    {
      flag_center = flag;
      flag_direction = flag;
      flag_radius = flag;
      flag_length_difference = flag;
      flag_cylinder = flag;
      flag_translate = flag;
    }


  // ********************************************************
  // MACROS FOR TEMPLATE CLASS FIELD_PARAM_T MEMBER FUNCTIONS
  // ********************************************************

#define _FIELD_PARAM_T_HEADER_                                       \
  template <typename DIM_TYPE, typename AXIS_SIZE_TYPE,              \
            typename COORD_TYPE, typename DIR_TYPE,                  \
            typename RADIUS_TYPE, typename DIFF_TYPE,                \
            typename ANGLE_TYPE, typename SCALAR_TYPE,               \
            typename SEED_TYPE,                                      \
            typename MIN_MAX_TYPE, typename NUM_TYPE>                \
  void FIELD_PARAM_T<DIM_TYPE, AXIS_SIZE_TYPE, COORD_TYPE, DIR_TYPE, \
                     RADIUS_TYPE, DIFF_TYPE, ANGLE_TYPE,             \
                     SCALAR_TYPE, SEED_TYPE,                         \
                     MIN_MAX_TYPE, NUM_TYPE>::

#define _GET_COORDINATE_STRING_(_function_name,_obj)                 \
  _FIELD_PARAM_T_HEADER_ _function_name                              \
  (const NUM_TYPE i, std::string & s) const                          \
  {                                                                  \
    convert_to_string(_obj.begin()+i*Dimension(),                    \
                      _obj.begin()+(i+1)*Dimension(), s);            \
  }

#define _GET_SCALAR_STRING_(_function_name,_obj)                     \
  _FIELD_PARAM_T_HEADER_ _function_name                              \
  (const NUM_TYPE i, std::string & s) const                          \
  {                                                                  \
    convert_to_string(_obj.begin()+i,                                \
                      _obj.begin()+i+1, s);                          \
  }

  // ********************************************************
  // TEMPLATE CLASS FIELD_PARAM_T MEMBER FUNCTIONS
  // ********************************************************

  template <typename T> 
  void convert_to_string(const T x, std::string & s)
  {
    std::ostringstream ostr("");

    ostr << x;
    s = ostr.str();
  }
    
	
  template <typename ITYPE> 
  void convert_to_string(const ITYPE list_begin, const ITYPE list_end,
                         std::string & s)
  {
      std::ostringstream ostr("");

      for (ITYPE iter = list_begin; iter != list_end; iter++) {
        ostr << *iter;
        if (iter+1 != list_end) { ostr << " "; }
      }
      s = ostr.str();
  }

  _FIELD_PARAM_T_HEADER_
  Init()
  {
    flag_tilt = true;
    flag_multi = false;
    flag_set_boundary2zero = false;
    gradient_discontinuity_zero = true;
  }

  _FIELD_PARAM_T_HEADER_
  SetAxisSize(const AXIS_SIZE_TYPE size)
  {
    IJK::PROCEDURE_ERROR error("FIELD_PARAM_T.SetAxisSIze");

    if (!dimension.IsSet()) {
      error.AddMessage
        ("Programming error. Dimension must be set before axis size.");
      throw error;
    }

    for (DIM_TYPE d = 0; d < Dimension(); d++) 
      { axis_size.push_back(size); }
  }

  _GET_COORDINATE_STRING_(GetCenterString, center);
  _GET_COORDINATE_STRING_(GetDirectionString, direction);
  _GET_COORDINATE_STRING_(GetSideDirectionString, side_direction);
  _GET_COORDINATE_STRING_(GetRotationDirectionString, rotation_direction);
  _GET_COORDINATE_STRING_(GetNormalString, normal);
  _GET_COORDINATE_STRING_(GetTranslateString, translate);

  _GET_SCALAR_STRING_(GetLengthDifferenceString, length_difference);
  _GET_SCALAR_STRING_(GetRadiusString, radius);
  _GET_SCALAR_STRING_(GetAngleString, angle);
};

#endif
