/// \file ijkgenscalarIO.cxx
/// Read/write/prompt for ijkgenscalar
/// Version v0.1.0

/*
  IJK: Isosurface Jeneration Code
  Copyright (C) 2011-2013 Rephael Wenger

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  (LGPL) as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <cmath>
#include <iomanip>
#include <iostream>

#include "ijkgenscalar.h"
#include "ijkgenscalarIO.h"

#include "ijkNrrd.h"
#include "ijkgrid_nrrd.txx"
#include "ijkstring.txx"

// Types
typedef IJK::NRRD_DATA<int, int> NRRD_HEADER;

using namespace IJKGENSCALAR;
using namespace std;

// **************************************************
// WRITE NRRD FILE
// **************************************************

void make_string
(const string & s_in, const int i, string & s_out)
{
  string s;

  IJK::val2string(i,s);
  s_out = s_in + s;
}

void add_nrrd_key_values
(const FIELD_PARAM & field_param, NRRD_HEADER & nrrd_header,
 const std::vector<FIELD_INFO> & field_info)
{
  const int field_index = field_param.FieldIndex();
  string field_name = field_info[field_index].name;

  if (field_info[field_index].flag_center) {

    if (field_param.NumCenters() == 1) {
      string center_string;
      field_param.GetCenterString(0, center_string);
      nrrdKeyValueAdd
        (nrrd_header.DataPtr(), "center", center_string.c_str());
    }
    else {
      for (int i = 0; i < field_param.NumCenters(); i++) {
        string key;
        string center_string;
        make_string("center", i, key);
        field_param.GetCenterString(i, center_string);
        nrrdKeyValueAdd
          (nrrd_header.DataPtr(), key.c_str(), center_string.c_str());
      }
    }
  }

  if (field_info[field_index].flag_cylinder) {

    if (field_param.NumDirections() == 1) {
      string coord_string;
      field_param.GetDirectionString(0, coord_string);
      nrrdKeyValueAdd
        (nrrd_header.DataPtr(), "cylinderAxisDirection", coord_string.c_str());
    }
  }

  if (field_name == "annulus" || field_name == "annulus_slice") {

    if (field_param.NumDirections() == 1) {
      string coord_string;
      field_param.GetDirectionString(0, coord_string);
      nrrdKeyValueAdd
        (nrrd_header.DataPtr(), "annulusAxisDirection", coord_string.c_str());
    }

  }

  if (field_name == "flange" || field_name == "flange_slice") {

    if (field_param.NumDirections() == 1) {
      string coord_string;
      field_param.GetDirectionString(0, coord_string);
      nrrdKeyValueAdd
        (nrrd_header.DataPtr(), "flangeAxisDirection", coord_string.c_str());
    }

  }

  if (field_name == "cube") {
    if (field_param.flag_tilt) {

      if (field_param.NumDirections() == 1) {
        string coord_string;
        field_param.GetDirectionString(0, coord_string);
        nrrdKeyValueAdd
          (nrrd_header.DataPtr(), "firstAxisDirection", coord_string.c_str());
      }

      if (field_param.NumSideDirections() == 1) {
        string coord_string;
        field_param.GetSideDirectionString(0, coord_string);
        nrrdKeyValueAdd
          (nrrd_header.DataPtr(), "secondAxisDirection", coord_string.c_str());
      }
    }
  }

  if (field_name == "constant_unit_gradient") {

    if (field_param.NumDirections() == 1) {
      string coord_string;
      field_param.GetDirectionString(0, coord_string);
      nrrdKeyValueAdd
        (nrrd_header.DataPtr(), "gradientDirection", coord_string.c_str());
    }
  }

  if (field_info[field_index].flag_radius) {

    if (field_param.NumRadii() == 1) {
      string radius_string;
      field_param.GetRadiusString(0, radius_string);
      nrrdKeyValueAdd
        (nrrd_header.DataPtr(), "radius", radius_string.c_str());
    }
  }

  if (field_info[field_index].flag_length_difference) {

    if (field_param.NumLengthDifferences() == 1) {
      string length_difference_string;
      field_param.GetLengthDifferenceString(0, length_difference_string);
      nrrdKeyValueAdd
        (nrrd_header.DataPtr(), "lengthDifference", 
         length_difference_string.c_str());
    }
  }

  if (field_name == "square") {

    if (field_param.NumSideDirections() == 1) {
      string coord_string;
      field_param.GetSideDirectionString(0, coord_string);
      nrrdKeyValueAdd
        (nrrd_header.DataPtr(), "sideDirection", coord_string.c_str());
    }

  }

  if (field_name == "edge") {

    if (field_param.NumAngles() == 1) {
      string angle_string;
      field_param.GetAngleString(0, angle_string);
      nrrdKeyValueAdd
        (nrrd_header.DataPtr(), "angle", angle_string.c_str());
    }

    if (field_param.NumNormals() == 1) {
      string normal_string;
      field_param.GetNormalString(0, normal_string);
      nrrdKeyValueAdd
        (nrrd_header.DataPtr(), "normal0", normal_string.c_str());
    }

    if (field_param.NumRotationDirections() == 1) {
      string direction_string;
      field_param.GetNormalString(0, direction_string);
      nrrdKeyValueAdd
        (nrrd_header.DataPtr(), "rotationDirection", direction_string.c_str());
    }

  }

  if (field_param.flag_stack) {

    if (field_param.NumTranslate() == 1) {
      string coord_string;
      field_param.GetTranslateString(0, coord_string);
      nrrdKeyValueAdd
        (nrrd_header.DataPtr(), "translate", coord_string.c_str());
    }

    string num_obj_str;
    make_string("", field_param.NumObjects(), num_obj_str);
    nrrdKeyValueAdd
      (nrrd_header.DataPtr(), "numStackObj", num_obj_str.c_str());

  }

  if (field_info[field_index].flag_random_seed) {
    string seed_string;
    IJKGENSCALAR::convert_to_string(field_param.RandomSeed(), seed_string);
    nrrdKeyValueAdd
      (nrrd_header.DataPtr(), "randomSeed", seed_string.c_str());
  }

  if (field_info[field_index].flag_maxval) {
    string maxval_string;
    IJKGENSCALAR::convert_to_string(field_param.MaxVal(), maxval_string);
    nrrdKeyValueAdd
      (nrrd_header.DataPtr(), "maxValue", maxval_string.c_str());
  }
}


void IJKGENSCALAR::write_scalar_grid
(const char * output_filename, const SCALAR_GRID & grid, 
 const FIELD_PARAM & field_param,  const std::vector<FIELD_INFO> & field_info)
{
  const int dimension = grid.Dimension();
  const int field_index = field_param.FieldIndex();
  NRRD_HEADER nrrd_header;
  IJK::ARRAY<double> grid_spacing(dimension, 1);

  for (int d = 0; d < dimension; d++) 
    { grid_spacing[d] = grid.Spacing(d); }

  nrrd_header.SetSize(grid.Dimension(), grid.AxisSize());
  nrrdAxisInfoSet_nva(nrrd_header.DataPtr(), nrrdAxisInfoSpacing, 
                      grid_spacing.PtrConst());
  nrrdKeyValueAdd
    (nrrd_header.DataPtr(), "ijkDatasetType",
     field_info[field_index].name.c_str());

  add_nrrd_key_values(field_param, nrrd_header, field_info);

  write_scalar_grid_nrrd(output_filename, grid, nrrd_header);
}

void IJKGENSCALAR::write_gradient_grid
(const string & output_filename, const GRADIENT_GRID & grid,
 const FIELD_PARAM & field_param, const std::vector<FIELD_INFO> & field_info)
{
  const int dimension = grid.Dimension();
  const int field_index = field_param.FieldIndex();
  const char * DISCONTINUITY_KEY = "handleGradientDiscontinuity";
  NRRD_HEADER nrrd_header;
  IJK::ARRAY<int> nrrd_axis_size(dimension+1);
  IJK::ARRAY<double> grid_spacing(dimension+1, 1);
  
  nrrd_axis_size[0] = grid.VectorLength();
  for (int d = 1; d <= dimension; d++) {
    nrrd_axis_size[d] = grid.AxisSize(d-1); 
    grid_spacing[d] = grid.Spacing(d-1);
  }

  nrrd_header.SetSize(dimension+1, nrrd_axis_size.PtrConst());
  nrrdAxisInfoSet_nva(nrrd_header.DataPtr(), nrrdAxisInfoSpacing, 
                      grid_spacing.PtrConst());

  std::string dataset_type = field_info[field_index].name;
  dataset_type += "Gradients";

  nrrdKeyValueAdd
    (nrrd_header.DataPtr(), "ijkDatasetType", dataset_type.c_str());

  add_nrrd_key_values(field_param, nrrd_header, field_info);

  if (field_param.gradient_discontinuity_zero) {
    nrrdKeyValueAdd
      (nrrd_header.DataPtr(), DISCONTINUITY_KEY, "setToZero");
  }
  else {
    nrrdKeyValueAdd
      (nrrd_header.DataPtr(), DISCONTINUITY_KEY, "selectOne");
  }

  write_vector_grid_nrrd(output_filename, grid, nrrd_header);
}
